# Extended example inheritance and ABC using Shape/Square/Circle defined in the notes.
# Also includes the derived ABC Shape1d (which defines one concrete function) and
#  and Shape2d (which defines none). Both are used as classes derived from Shape,
#  from which the concrete classes (Square, Circle, Triangle, Line, Arc) are derived.

import math   # use pi
import abc    # use the class ABC and the function abstractmethod (a decorater)

class Shape(abc.ABC):
    @abc.abstractmethod
    def perimeter(self): pass  # will be overridden in some derived class

    @abc.abstractmethod
    def area(self): pass       # will be overridden in some derived class

    def one_dimensional(self): # concrete: works by calling an abstract method
        return self.area() == 0

class Shape2d(Shape):
    pass

class Square(Shape2d):
    def __init__(self,side)  : self.side = side
    def __repr__(self)       : return f'Square(side={self.side})'
    def perimeter(self)      : return 4*self.side
    def area(self)           : return self.side**2

class Circle(Shape2d):
    def __init__(self,radius): self.radius = radius
    def __repr__(self)       : return f'Circle(radius={self.radius})'
    def perimeter(self)      : return math.pi * 2*self.radius
    def area(self)           : return math.pi * self.radius**2

class Triangle(Shape2d):
    def __init__(self,a,b,c) : self.a, self.b, self.c = a, b, c
    def __repr__(self)       : return f'Triangle{self.a,self.b,self.c}'
    def perimeter(self)      : return self.a + self.b + self.c
    def area(self):
        sp = self.perimeter()/2
        return math.sqrt(sp*(sp-self.a)*(sp-self.b)*(sp-self.c))

class Shape1d(Shape):
    def area(self) : return 0

class Line(Shape1d):
    def __init__(self,length): self.length = length
    def __repr__(self)       : return f'Line(length={self.length})'
    def perimeter(self)      : return self.length

class Arc(Shape1d):
    def __init__(self,radius,angle): self.radius, self.angle = radius, angle
    def __repr__(self)       : return f'Arc(radius={self.radius},angle={self.angle})'
    def perimeter(self)      : return math.pi * 2*self.radius * self.angle/360



try:
    shape_try = Shape()
except:
    print('Exception raised: could not construct Shape object')

try:
    shape1d_try = Shape1d()
except:
    print('Exception raised: could not construct Shape1d object')

try:
    shape2d_try = Shape2d()
except:
    print('Exception raised: could not construct Shape2d object')

asquare = Square(2)
print(f'\n{asquare} perimieter={asquare.perimeter()} area={asquare.area()} 1d={asquare.one_dimensional()}')
print(f'isinstance(asquare,Square)={isinstance(asquare,Square)} isinstance(asquare,Shape2d)={isinstance(asquare,Shape2d)} isinstance(asquare,Shape)={isinstance(asquare,Shape)}')
acircle = Circle(2)
print(f'\n{acircle} perimieter={acircle.perimeter()} area={acircle.area()} 1d={acircle.one_dimensional()}')

atriangle = Triangle(3,4,5)
print(f'\n{atriangle} perimieter={atriangle.perimeter()} area={atriangle.area()} 1d={atriangle.one_dimensional()}')

anarc = Line(5)
print(f'\n{anarc} perimieter={anarc.perimeter()} area={anarc.area()} 1d={anarc.one_dimensional()}')

anarc = Arc(3,20)
print(f'\n{anarc} perimieter={anarc.perimeter()} area={anarc.area()} 1d={anarc.one_dimensional()}')
