# Non-negative property factory example defined in the notes.
# Here, anyone can bind the count property, but it must be bound to a
#  non-negative value.
# Note how the fix_property_names decorator function updates the
#  special_attr_name to be related to both the name of the attribute
#  (count), the name of the factor (non_negative), and a unique id.
  

def non_negative():
 
    # Each time non_negative is called, its id attribute is different.
    # Concatenate a speical attribute name using 'non_negative' and the id.
    if hasattr(non_negative,'id'):
        non_negative.id += 1
    else:
        non_negative.id = 0
    special_attr_name = 'non_negative#'+str(non_negative.id)

    def get_non_negative(an_object):
        return getattr(an_object,special_attr_name)
 
    def set_non_negative(an_object, value):
        if  value < 0:
            raise ValueError(f'non_negative.set_non_negative: value({value}) not >= 0')
        setattr(an_object,special_attr_name, value)
 
    def del_non_negative(an_object):
        delattr(an_object,special_attr_name)
 
    return property(get_non_negative, set_non_negative, del_non_negative)


def fix_property_names(aclass):
    for attr_name, value in aclass.__dict__.items():
        if isinstance(value,property)\
             and value.fset != None\
             and 'special_attr_name' in value.fset.__code__.co_freevars:
            cell_to_update = value.fset.__closure__[value.fset.__code__.co_freevars.index('special_attr_name')]
            cell_to_update.cell_contents = f'{attr_name}@{cell_to_update.cell_contents}'
    return aclass

@fix_property_names # Comment-out to not fix property names
class C:
    def __init__(self,init):
        self.count = init

    def inc(self):
        self.count += 1

    #...def other methods...

    count = non_negative()

o = C(5)
print('o = C(5)')
print('o namespace=',o.__dict__)
print('search for o.count=',o.count)

print('\no.inc()')
o.inc()
print('search for o.count=',o.count)

print('\no.count = 1')
o.count = 1
print('search for o.count=',o.count)

print('\ndel o.count')
del o.count
print('o namespace=',o.__dict__)

try:
    print("\nprint('search for o.count=',o.count)")
    print('search for o.count=',o.count) 
except:
    print('Exception raised: could not compute o.count')
